;;; howm-common.el --- Wiki-like keyword jump between files
;;; Copyright (c) 2002, 2003, 2004
;;;   by HIRAOKA Kazuyuki <khi@users.sourceforge.jp>
;;; $Id: howm-common.el,v 1.14 2004/01/23 11:48:59 hira Exp $
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 1, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; The GNU General Public License is available by anonymouse ftp from
;;; prep.ai.mit.edu in pub/gnu/COPYING.  Alternately, you can write to
;;; the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
;;; USA.
;;--------------------------------------------------------------------

(defvar howm-directory "~/howm/"
  "All files under this directory are scanned recursively.")

;; (defvar howm-file-name-format "%Y/%m/%Y_%m_%d.howm"
(defvar howm-file-name-format "%Y/%m/%Y-%m-%d-%H%M%S.howm"
  "Name of new file. See `format-time-string'.
You must guarantee (string< oldfile newfile).")

(defvar howm-abbreviate-file-name t)
(defun howm-abbreviate-file-name (f)
  (if (howm-abbreviate-file-name-p)
      (abbreviate-file-name f)
    f))
;; for compatibility (incomplete)
(defun howm-abbreviate-file-name-p ()
  (cond ((boundp 'howm-template-file-abbrev)
         (progn
           (defvar howm-template-file-abbrev nil) ;; Inhibit warning. Sigh...
           howm-template-file-abbrev))
        ((boundp 'howm-view-header-abbrev)
         (progn
           (defvar howm-view-header-abbrev nil) ;; Inhibit warning. Sigh...
           howm-view-header-abbrev))
        (t howm-abbreviate-file-name)))
;; (defun howm-abbreviate-file-name-p () howm-abbreviate-file-name)

;;; for XEmacs fallback
(if (not (fboundp 'font-lock-fontify-block))
    (defalias 'font-lock-fontify-block 'font-lock-fontify-buffer))
;;; If you use APEL, you can replace a below block with (require 'poe).
(if (not (fboundp 'line-beginning-position))
    (defalias 'line-beginning-position 'point-at-bol))
(if (not (fboundp 'line-end-position))
    (defalias 'line-end-position 'point-at-eol))
;;; Imported from APEL 10.6
(if (not (fboundp 'match-string-no-properties))
    ;; Emacs 20.3 and later: (match-string-no-properties NUM &optional STRING)
    (defun match-string-no-properties (num &optional string)
      "Return string of text matched by last search, without text properties.
NUM specifies which parenthesized expression in the last regexp.
 Value is nil if NUMth pair didn't match, or there were less than NUM pairs.
Zero means the entire text matched by the whole regexp or whole string.
STRING should be given if the last search was by `string-match' on STRING."
      (if (match-beginning num)
          (if string
              (let ((result
                     (substring string (match-beginning num) (match-end num))))
                (set-text-properties 0 (length result) nil result)
                result)
            (buffer-substring-no-properties (match-beginning num)
                                            (match-end num))))))

(defvar howm-message-time nil)
(defun howm-message-time-from (ti0)
  (when howm-message-time
    (let ((ti1 (current-time)))
      (message "took %.2f sec."
               (howm-time-difference-second ti1 ti0)))))
(defun howm-time-difference-second (ti1 ti0)
  (let ((h (- (car ti1) (car ti0)))
        (l (- (second ti1) (second ti0)))
        (m (- (third ti1) (third ti0))))
    (+ (* h 65536) l (* m 1e-6))))

(defun howm-xor (a b)
  (if a (not b) b))
(defun howm-buffer-empty-p (&optional buf)
  ;; emacs20 doesn't have (buffer-size buf)
  (with-current-buffer buf
    (= (buffer-size) 0)))
;; (defun howm-buffer-empty-p (&optional buf)
;;   (= (buffer-size buf) 0))
(defun howm-point-beginning-of-line ()
  (save-excursion
    (beginning-of-line)
    (point)))
(defun howm-point-end-of-line ()
  (save-excursion
    (end-of-line)
    (point)))

(defun howm-inhibit-warning-in-compilation (&rest dummy)
  (error "This is dummy function to inhibit warning in compilation."))

(defvar howm-view-font-lock-silent t
  "Inhibit font-lock-verbose if non-nil.")

;; (defvar howm-reminder-quick-check-key ";")

;;; cl

;; (defmacro howm-define-maybe (fname fargs &rest fbody)
;;   (when (not (fboundp fname))
;;     `(defun ,fname ,fargs
;;        ,@fbody)))

;; (howm-define-maybe caddr (x)
;;                    (car (cdr (cdr x))))

;; (howm-define-maybe second (x)
;;                    (cadr x))

;; (howm-define-maybe third (x)
;;                    (caddr x))

;;;

(provide 'howm-common)

;;; howm-common.el ends here
