;;; howm-menu.el --- Wiki-like keyword jump between files
;;; -*- Coding: iso-2022-7bit -*-
;;; Copyright (c) 2002, 2003, 2004
;;;   by HIRAOKA Kazuyuki <khi@users.sourceforge.jp>
;;; $Id: howm-menu.el,v 1.38 2004/02/06 11:13:55 hira Exp $
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 1, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; The GNU General Public License is available by anonymouse ftp from
;;; prep.ai.mit.edu in pub/gnu/COPYING.  Alternately, you can write to
;;; the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
;;; USA.
;;;--------------------------------------------------------------------

(defvar howm-menu-key-face 'howm-menu-key-face)
(defface howm-menu-key-face ;; shortcut key in menu-mode
  '((((class color) (background light)) (:foreground "dark red"))
    (((class color) (background dark)) (:foreground "orange")))
  "Face for `howm-menu-mode'")

(defun howm-menu-refresh-background ()
  (let ((b (current-buffer)))
    (howm-menu t)
    (switch-to-buffer b)))

(defun howm-open-today ()
  (interactive)
  (and (howm-create-file t)
       (howm-insert-template ""))
  (howm-set-mode))

(defun howm-open-past (&optional days-before)
  (interactive "p")
  (let ((f (expand-file-name (howm-file-name (howm-days-after (current-time)
                                                              (- days-before)))
                             howm-directory)))
    (if (file-exists-p f)
        (find-file f)
      (error "No such file: %s" f)))
  (howm-set-mode))

;; (defun howm-open-past (&optional days-before)
;;   (interactive)
;;   (if (= 0 (or days-before 0))
;;       (howm-create-file t)
;;     (find-file
;;      (expand-file-name (howm-file-name (howm-days-after (current-time)
;;                                                         (- days-before)))
;;                        howm-directory)))
;;   (howm-set-mode))

(defun howm-menu-eval (s)
  (let ((expr (read s)))
    (eval expr)))
(defun howm-menu-call (s)
  (let ((expr (read s)))
    (call-interactively expr)))

(defvar howm-menu-toggle-invisible nil)
;; (defvar howm-menu-toggle-invisible "%|")
(defun howm-menu-make-invisible ()
  (save-excursion
    (goto-char (point-min))
    (let (visible-p
          invisible-beg)
      (while (not (= (point) (point-max)))
        (setq visible-p t)
        (while (re-search-forward howm-menu-toggle-invisible
                                  (line-end-position) t)
          (if visible-p
              (setq invisible-beg (match-beginning 0))
            (put-text-property invisible-beg (match-end 0) 'invisible t))
          (setq visible-p (not visible-p)))
        (when (not visible-p)
          (put-text-property invisible-beg
                             (save-excursion (forward-line) (point))
                             'invisible t))
        (forward-line)))))

;; howm-menu-command-table-* = ((MATCHER FUNC ONBUF) ...)
;; 
;; (FUNC) is evalueted on ONBUF when return key is hit on MATCHER.
;; 
;; MATCHER = regexp | (regexp position)
;; (optional) ONBUF = nil | 'previous | 'current
;;   nil: previous non-menu buffer (set-buffer)
;;   'previous: previous non-menu buffer (switch-to-buffer)
;;   'current: current menu buffer

(defvar howm-menu-action-arg 'howm-menu-action-arg-name)
(defvar howm-menu-command-table-common
  '(
    (("%eval%\\(.*$\\)" 1) howm-menu-eval previous)
    (("%call%\\(.*$\\)" 1) howm-menu-call previous)
     ))
(defvar howm-menu-command-table-en
  `(
    ("[New]" (lambda () (howm-create ,howm-menu-action-arg)))
;;     ("[New]" howm-create)
    ("[Dup]" howm-dup)
    ("[Update]" howm-initialize-buffer previous)
    ("[Regexp]" howm-list-grep)
    ("[String]" howm-list-grep-fixed)
    ;;         ("[roma]" howm-list-migemo)
    ("[Today]" howm-open-today)
    ("[Yesterday]" (lambda () (howm-open-past 1)))
    ("[All]" howm-list-all)
    ("[Recent]" howm-list-recent)
    ("[Around]" howm-list-around)
    ("[Schedule]" howm-list-schedule)
    ("[<Title]" howm-keyword-to-kill-ring)
    ("[<Name]" (lambda () (howm-keyword-to-kill-ring t)))
    ("[Key>]" howm-insert-keyword previous)
    ("[Date>]" howm-insert-date previous)
    ("[Todo]" howm-list-todo)
    ("[Killall]" howm-kill-all)
    ("[Force Killall]" (lambda () (interactive) (howm-kill-all t)))
    ("[Edit Menu]" howm-menu-edit current)
    ("[Update Menu]" howm-menu-refresh current)
    ))
(defvar howm-menu-command-table-ja
  `(
    ;; Japanese characters below
    ("[$B?75,(B]" (lambda () (howm-create ,howm-menu-action-arg)))
;;     ("[$B?75,(B]" howm-create)
    ("[$BJ#@=(B]" howm-dup)
    ("[$B99?7(B]" howm-initialize-buffer previous)
    ("[$B@55,(B]" howm-list-grep)
    ("[$B8GDj(B]" howm-list-grep-fixed)
    ("[roma]" howm-list-migemo)
    ("[$B:#F|(B]" howm-open-today)
    ("[$B:rF|(B]" (lambda () (howm-open-past 1)))
    ("[$B0lMw(B]" howm-list-all)
    ("[$B:G6a(B]" howm-list-recent)
    ("[$BA08e(B]" howm-list-around)
    ("[$BM=Dj(B]" howm-list-schedule)
    ("[$BBj",(B]" howm-keyword-to-kill-ring)
    ("[$BL>",(B]" (lambda () (howm-keyword-to-kill-ring t)))
    ("[$B80"-(B]" howm-insert-keyword previous)
    ("[$BF|"-(B]" howm-insert-date previous)
    ("[Todo]" howm-list-todo)
    ("[$BA4>C(B]" howm-kill-all)
    ("[$B6/@)A4>C(B]" (lambda () (interactive) (howm-kill-all t)))
    ("[menu $BJT=8(B]" howm-menu-edit current)
    ("[menu $B99?7(B]" howm-menu-refresh current)
    ))
(defun howm-menu-action-lock-rules ()
  (let* ((d action-lock-default-rules)
         (f (howm-action-lock-reminder-forward-rules))
         (reg-head "^\\(>\\)[^0-9\r\n]*[0-9]+")
         (reg (if howm-menu-todo-priority-format  ;; for fear of efficiency
                  (concat reg-head
                          (format ".* +\\(%s.*\\)$"
                                  (howm-reminder-regexp howm-reminder-types)))
                (concat reg-head " +\\(.*\\)$")))
         (j (list (howm-action-lock-search reg 2 1 nil t)))
;;         (j (list (howm-action-lock-search "^\\(>\\)[^0-9\r\n]*[0-9]+ +\\(.*\\)$"
;;                                           2 1)))
        (m (mapcar (lambda (pair)
                     (let* ((h (car pair))
                            (r (if (listp h) (car h) h))
                            (n (if (listp h) (second h) nil))
                            (args (if n
                                     `(list (match-string-no-properties ,n))
                                    nil))
                            (functab (cdr pair))
                            (c (howm-menu-action functab args)))
                       (list r c)))
                   (howm-menu-command-table))))
    (append m d j f)))

;; Elisp is not Scheme. Lambda is not closure. Don't forget dynamic binding.
;; Check
;;   (pp (car (howm-menu-action-lock-rules)))
;; for debug. [2003/09/25]
(defun howm-menu-action (function-table args)
  (let* ((func (car function-table))
         (onbuf (second function-table))
         (switch-p (eq onbuf 'previous)))
    (let* ((s-buf (if (eq onbuf 'current) 'cur 'prev))
           (s-switch `(switch-to-buffer ,s-buf))
           (s-apply `(apply #',func ,(if args 'a nil))))
;;            (s-apply `(apply #',func ,(if args '(list a) nil))))
      (let* ((s-body (if switch-p
                         `(progn ,s-switch ,s-apply)
                       `(with-current-buffer ,s-buf ,s-apply))))
        `(lambda (&optional ,howm-menu-action-arg)
           (let ((a ,args)
                 (cur (current-buffer))
                 (prev (if (howm-buffer-alive-p howm-menu-previous-buffer)
                           howm-menu-previous-buffer
                         (current-buffer))))
             ,s-body))))))

(defun howm-buffer-alive-p (buf)
  (and buf (buffer-name buf)))

(defvar howm-menu-file nil)
;; (defvar howm-menu-file "menu.howm")
(defvar howm-menu-name-format "*howmM:%s*")
(defvar howm-menu-schedule-days 7)
(defvar howm-menu-schedule-days-before 0)
(defvar howm-menu-todo-num 50)
(defvar howm-menu-todo-priority -66666)
(defvar howm-menu-expiry-hours 0)
(defvar howm-menu-hook nil)

;; Trying...
;; (defvar howm-menu-keyword-regexp nil)
;; (defvar howm-menu-top nil)
(defvar howm-menu-keyword-regexp "^%.*%$")
(defvar howm-menu-top "%menu%")

;; internal
(defvar howm-menu-previous-buffer nil)
(defvar howm-menu-next-expiry-time (current-time))
(defvar howm-menu-last-time (current-time))
(defvar howm-menu-buffer-file nil)
(defvar howm-menu-buffer-file-place nil)
(defvar howm-menu-mode-map nil)
(make-variable-buffer-local 'howm-menu-previous-buffer)
(make-variable-buffer-local 'howm-menu-next-expiry-time)
(make-variable-buffer-local 'howm-menu-last-time)
(make-variable-buffer-local 'howm-menu-buffer-file)
(make-variable-buffer-local 'howm-menu-buffer-file-place)
(make-variable-buffer-local 'howm-menu-mode-map)

(defvar howm-day-of-week-en "SMTWtFs")
;; Japanese characters below
(defvar howm-day-of-week-ja "$BF|7n2P?eLZ6bEZ(B")

(defvar howm-menu-lang
  (if (or (featurep 'wnn)
          (featurep 'canna)
          (featurep 'skk)
          (and (boundp 'current-language-environment)
               (string= current-language-environment "Japanese")))
      'ja
    'en))
(defun howm-menu-lang-assoc (lang)
  (let ((a `((default . (,howm-menu-command-table-en
;;                          ,howm-menu-header-en
                         ,howm-day-of-week-en))
             (ja . (,howm-menu-command-table-ja
;;                     ,howm-menu-header-ja
                    ,howm-day-of-week-ja)))))
    (cdr (or (assoc lang a)
             (assoc 'default a)))))
(defun howm-menu-command-table-raw ()
  (car (howm-menu-lang-assoc howm-menu-lang)))
(defun howm-menu-command-table ()
  (append howm-menu-command-table-common
          (mapcar (lambda (pair) (cons (regexp-quote (car pair)) (cdr pair)))
                  (howm-menu-command-table-raw))))
;; (defun howm-menu-header ()
;;   (second (howm-menu-lang-assoc howm-menu-lang)))
(defun howm-day-of-week ()
  (second (howm-menu-lang-assoc howm-menu-lang)))

(defun howm-time< (t1 t2)
  (or (< (car t1) (car t2))
      (and (= (car t1) (car t2))
           (< (second t1) (second t2)))))

(defvar *howm-menu-force-refresh* nil) ;; dirty. clean me. [2003/09/29 21:39]
(defun howm-menu (&optional force-refresh last-chance)
  (interactive)
  (when (not (file-exists-p howm-directory))
    (make-directory howm-directory t))
  (let ((*howm-menu-force-refresh* force-refresh)
        ;; force to use the original howm-directory
        (*howm-independent-directories* nil))
    (if (and howm-menu-keyword-regexp howm-menu-top)
        (let ((m (howm-keyword-search howm-menu-top)))
          (when (and (cdr (assoc 'menu-p m))
                     (not (cdr (assoc 'keyword-matched m))))
            (howm-menu-initialize-skel last-chance)))
      (howm-menu-open howm-menu-file))))
;; (defun howm-menu (&optional force-refresh)
;;   (interactive)
;;   (let ((*howm-menu-force-refresh* force-refresh))
;;     (if (and howm-menu-keyword-regexp howm-menu-top)
;;         (howm-keyword-search howm-menu-top)
;;       (howm-menu-open howm-menu-file))))

(defun howm-menu-initialize-skel (no-chance)
  (let ((fd (howm-menu-skel)))
    (if (and fd (not no-chance))
        (progn
          (howm-menu-copy-skel fd)
          (howm-view-kill-buffer)
          (howm-menu nil t))
      (error "Can't find menu-file."))))

(defun howm-menu-skel ()
  (let* ((v (if (eq howm-menu-lang 'ja)
                'howm-ja-dir
              'howm-en-dir))
         (d (and (boundp v) (eval v)))
         (f "0000-00-00-000000.howm"))
    (if (and d (file-exists-p (expand-file-name f d)))
        (list f d)
      nil)))

(defun howm-menu-copy-skel (file-dir &optional menu-file)
  (let* ((f (car file-dir))
         (d (second file-dir))
         (s (expand-file-name f d))
         (menu-file (or menu-file 
                        (expand-file-name f howm-directory))))
    (copy-file s menu-file)))

(defun howm-menu-open (file &optional place name)
  (setq name (or name (howm-menu-name file)))
  (let ((f (expand-file-name file howm-directory)))
    (if (file-exists-p f)
        (let* ((pb (current-buffer))
               (pm major-mode)
               (b (get-buffer name))
               (m (nth 5 (file-attributes f))))
          (if (or *howm-menu-force-refresh*
                  (null b)
                  (progn
                    (set-buffer b)
                    (or (howm-time< howm-menu-last-time m)
                        (howm-time< howm-menu-next-expiry-time
                                    (current-time)))))
              (howm-menu-refresh file place name)
            (switch-to-buffer b))
          (let ((cm major-mode))
            (save-excursion
              (while (eq pm cm)
                (set-buffer pb)
                (setq pb howm-menu-previous-buffer)
                (set-buffer pb)
                (setq pm major-mode)))
            (setq howm-menu-previous-buffer pb)))
;;           (when (not (eq pm major-mode))
;;             (setq howm-menu-previous-buffer pb)))
      (progn
        (find-file f)
        (howm-mode)))))

(defvar howm-menu-key-regexp
  "%\"\\(\\([^\r\n%\"]\\)[^\r\n%\"]*\\(%+[^\r\n%\"]+\\)*\\)\\(%\\)?\"")
(defvar howm-menu-key-regexp-word-pos 1)
(defvar howm-menu-key-regexp-key-pos 2)
(defvar howm-menu-key-regexp-moveonly-pos 4)
(defun howm-menu-add-font-lock ()
  (let ((ks `((,howm-menu-key-regexp
               (,howm-menu-key-regexp-key-pos howm-menu-key-face t)))))
    (cheat-font-lock-append-keywords ks)))

(defvar howm-menu-display-rules
  `(
    ("%sdays" . ,(lambda ()
                   (howm-menu-display-number howm-menu-schedule-days)))
    ("%tnum" . ,(lambda ()
                  (howm-menu-display-number howm-menu-todo-num)))
    ("%schedule[\n]?" . howm-menu-schedule)
    ("%todo[\n]?" . howm-menu-todo)
    (,howm-menu-key-regexp . howm-menu-shortcut)
  ))

(defun howm-menu-display-string (str)
  (delete-region (match-beginning 0) (match-end 0))
  (insert str))
(defun howm-menu-display-number (num)
  (howm-menu-display-string (format "%d" num)))

(defvar *howm-menu-shortcut-keys* nil)  ;; internal use
(defvar *howm-menu-shortcut-multidef-keys* nil)  ;; internal use
(defvar *howm-menu-shortcut-markers* nil)  ;; internal use
(make-variable-buffer-local '*howm-menu-shortcut-markers*)
(defun howm-menu-shortcut-get-marker ()
  (let ((m (make-marker)))
    (set-marker m (point))
    (add-to-list '*howm-menu-shortcut-markers* m)
    m))
(defun howm-menu-shortcut-clear-markers ()
  (mapc (lambda (m) (set-marker m nil))
        *howm-menu-shortcut-markers*)
  (setq *howm-menu-shortcut-markers* nil))
(defun howm-menu-shortcut-initialize ()
  (setq *howm-menu-shortcut-keys* nil)
  (setq *howm-menu-shortcut-multidef-keys* nil)
  (howm-menu-shortcut-clear-markers))
(defun howm-menu-shortcut-warn ()
  (when *howm-menu-shortcut-multidef-keys*
    (beep)
    (message "Multiple definitions for key(s): %s"
             *howm-menu-shortcut-multidef-keys*)))
;; Check howm-menu-mode-map for debug
(defun howm-menu-shortcut ()
  (let* ((beg (match-beginning 0))
         (end (match-end 0))
         (wbeg (match-beginning howm-menu-key-regexp-word-pos))
         (wend (match-end  howm-menu-key-regexp-word-pos))
;;          (word (match-string-no-properties howm-menu-key-regexp-word-pos))
         (key (match-string-no-properties howm-menu-key-regexp-key-pos))
;;         (key (substring word 0 1)) ;; first letter
         (move-only (match-beginning howm-menu-key-regexp-moveonly-pos)))
    (put-text-property beg wbeg 'invisible t)
    (put-text-property wend end 'invisible t)
;;     (delete-region beg end)
;;     (insert label)
    (let ((p (howm-menu-shortcut-get-marker)))
      (define-key howm-menu-mode-map key
        (howm-menu-shortcut-func p move-only)))
    (when (member key *howm-menu-shortcut-keys*)
      (setq *howm-menu-shortcut-multidef-keys*
            (cons key *howm-menu-shortcut-multidef-keys*)))
    (setq *howm-menu-shortcut-keys*
          (cons key *howm-menu-shortcut-keys*))))
(defun howm-menu-shortcut-func (p move-only)
  `(lambda (arg)
     (interactive "P")
     (let ((pos ,p))
       (if ,move-only
           (goto-char pos)
         (save-excursion
           (goto-char pos)
           (let ((case-fold-search nil)) ;; temporaly
             (when (null (action-lock-get-action))
               (action-lock-goto-next-link))
             (action-lock-invoke arg)))))))

(defun howm-menu-schedule ()
  (message "scanning schedule...")
  (delete-region (match-beginning 0) (match-end 0))
  (mapc #'howm-menu-insert-item
        (howm-schedule-menu howm-menu-schedule-days
                            howm-menu-schedule-days-before))
  (message "scanning schedule...done")
  )
(defun howm-menu-todo ()
  (message "scanning todo...")
  (delete-region (match-beginning 0) (match-end 0))
  (mapc (lambda (item) (howm-menu-insert-item item "  " t))
        (howm-todo-menu howm-menu-todo-num howm-menu-todo-priority))
  (message "scanning todo...done")
  )

(defun howm-menu-replace (rules &optional beg end)
  (mapc (lambda (pair)
          (let ((spell (car pair))
                (disp-f (cdr pair)))
            (goto-char (or beg (point-min)))
            (while (re-search-forward spell end t)
              (funcall disp-f))))
        rules))

(defun howm-menu-name (file)
  (format howm-menu-name-format file))

(defvar howm-menu-footer nil)
(defun howm-menu-footer ()
  (or howm-menu-footer
      (let* ((r (howm-menu-command-table-raw))
             (buttons (mapcar (lambda (f)
                                (cdr (assoc f
                                            (mapcar (lambda (z)
                                                      (cons (second z)
                                                            (car z)))
                                                    r))))
                              '(howm-menu-refresh howm-menu-edit)))
             (footer (apply #'concat `("\n-- \n" ,@buttons))))
        (setq howm-menu-footer footer)
        footer)))

(defun howm-menu-refresh (&optional file place name)
  (interactive)
  (when name
    (switch-to-buffer (get-buffer-create name)))
;;   (setq name (or name *howm-menu-last-name*))
;;   (switch-to-buffer (get-buffer-create name))
  (howm-menu-mode)
  (setq howm-menu-buffer-file (or file howm-menu-buffer-file))
  (setq buffer-read-only nil)
  (erase-buffer)
  (insert-file-contents (expand-file-name howm-menu-buffer-file
                                          howm-directory))
  (setq howm-menu-buffer-file-place (or place
                                      howm-menu-buffer-file-place
                                      1))
  (howm-view-set-place howm-menu-buffer-file-place)
  (let* ((r (howm-view-paragraph-region))
         (b (car r))
         (e (second r)))
    (delete-region e (point-max))
    (delete-region (point-min) b))
  (goto-char (point-max))
  (insert (howm-menu-footer))
  (howm-menu-shortcut-initialize)
  (howm-menu-replace howm-menu-display-rules)
;;   (let ((howm-view-excluded-file-regexp howm-excluded-file-regexp)) ;; dirty!
;;     (howm-menu-replace howm-menu-display-rules))
;;   (let* ((action-lock-default-rules (howm-menu-action-lock-rules)))
;;     (howm-mode 1))
  (let* ((action-lock-default-rules (howm-menu-action-lock-rules)))
    (if howm-mode
        (howm-initialize-buffer)
      (howm-mode 1)))
  (make-local-variable 'font-lock-keywords-only)
  (setq font-lock-keywords-only t)
  (howm-menu-add-font-lock)
  (font-lock-fontify-buffer)
  (when howm-menu-toggle-invisible
    (howm-menu-make-invisible))
  (set-buffer-modified-p nil)
  (setq buffer-read-only t)
  (goto-char (point-min))
  (setq howm-menu-last-time (current-time))
  (setq howm-menu-next-expiry-time
        (howm-days-after (current-time) 0
                         howm-menu-expiry-hours))
  (howm-menu-shortcut-warn)
  (run-hooks 'howm-menu-hook))

(defun howm-menu-edit ()
  (interactive)
  (let ((place howm-menu-buffer-file-place))
    (find-file (expand-file-name howm-menu-buffer-file howm-directory))
;;   (find-file (expand-file-name *howm-menu-last-file* (howm-directory)))
;;   (find-file (expand-file-name howm-menu-file (howm-directory)))
    (howm-mode t)
    (when place
      (howm-view-set-place place)
      (recenter 0))))

(defvar howm-menu-todo-priority-format nil)
;(defvar howm-menu-todo-priority-format "(%8.1f)")
(defun howm-menu-insert-item (item &optional day-of-week-str show-priority)
  (let* ((p (howm-todo-parse item))
         (late (floor (car p)))
         (dow (fourth p))
         (dow-str (or day-of-week-str
                      (howm-day-of-week-string dow)))
         (priority (if (and howm-menu-todo-priority-format
                            show-priority)
                       (format howm-menu-todo-priority-format
                               (howm-todo-priority item))
                     ""))
         (h (format "> %s%3s%s " dow-str late priority)))
    (insert h (howm-view-item-summary item) "\n")))

(defun howm-day-of-week-string (&optional day-of-week)
  ;; 0 = Sunday
  (let ((dow (or day-of-week (nth 6 (decode-time)))))
    (substring (howm-day-of-week) dow (1+ dow))))

(defun howm-menu-p ()
  (string= major-mode "howm-menu-mode"))

(defun howm-menu-mode ()
  "howm menu
key	binding
---	-------
\\[action-lock-magic-return]	Follow link
\\[action-lock-goto-next-link]	Next link
\\[action-lock-goto-previous-link]	Prev link
\\[describe-mode]	This help
\\[bury-buffer]	Quit
"
  (interactive)
  (setq major-mode 'howm-menu-mode
        mode-name "HM")
  (let ((m (make-keymap)))
    (define-key m [tab] 'action-lock-goto-next-link)
    (define-key m [(meta tab)] 'action-lock-goto-previous-link)
    (define-key m " " 'scroll-up)
    (define-key m [backspace] 'scroll-down)
    (define-key m "\C-h" 'scroll-down)
    (define-key m "q" 'bury-buffer)
    (define-key m "?" 'describe-mode)
;;     (define-key m howm-reminder-quick-check-key 'howm-reminder-quick-check)
    (setq howm-menu-mode-map m))
  (use-local-map howm-menu-mode-map)
  )

(provide 'howm-menu)

;;; howm-menu.el ends here
